const std = @import("std");

// Convenience fns init from file

pub fn GeoJsonZig(comptime N: usize, comptime T: type) type {
    return struct {
        const Self = @This();

        pub const Position = [N]T;

        pub const GeoJsonType = enum {
            Feature,
            FeatureCollection,
            Geometry,
        };

        pub const GeometryType = enum {
            Point,
            MultiPoint,
            LineString,
            MultiLineString,
            Polygon,
            MultiPolygon,
        };

        pub const Coordinates = union(GeometryType) {
            Point: Position,
            MultiPoint: []Position,
            LineString: []Position,
            MultiLineString: [][]Position,
            Polygon: [][]Position,
            MultiPolygon: [][][]Position,
        };

        pub const Geometry = struct {
            type: GeometryType,
            coordinates: Coordinates,
            allocator: std.mem.Allocator,

            pub fn initPoint(coordinates: Position) Geometry {
                return .{ .type = .Point, .coordinates = .{ .Point = coordinates }, .allocator = undefined };
            }

            pub fn initMultiPoint(allocator: std.mem.Allocator, coordinates: []const Position) !Geometry {
                const coords = try allocator.dupe(Position, coordinates);
                return .{ .type = .MultiPoint, .coordinates = .{ .MultiPoint = coords }, .allocator = allocator };
            }

            pub fn initLineString(allocator: std.mem.Allocator, coordinates: []const Position) !Geometry {
                const coords = try allocator.dupe(Position, coordinates);
                return .{ .type = .LineString, .coordinates = .{ .LineString = coords }, .allocator = allocator };
            }

            pub fn initMultiLineString(allocator: std.mem.Allocator, coordinates: []const []const Position) !Geometry {
                const coords = try allocator.alloc([]Position, coordinates.len);
                errdefer allocator.free(coords);
                for (coordinates, 0..) |line, i| {
                    coords[i] = try allocator.dupe(Position, line);
                    errdefer {
                        for (coords[0..i]) |c| {
                            allocator.free(c);
                        }
                        allocator.free(coords);
                    }
                }
                return .{ .type = .MultiLineString, .coordinates = .{ .MultiLineString = coords }, .allocator = allocator };
            }

            pub fn initPolygon(allocator: std.mem.Allocator, coordinates: []const []const Position) !Geometry {
                const coords = try allocator.alloc([]Position, coordinates.len);
                errdefer allocator.free(coords);
                for (coordinates, 0..) |line, i| {
                    coords[i] = try allocator.dupe(Position, line);
                    errdefer {
                        for (coords[0..i]) |c| {
                            allocator.free(c);
                        }
                        allocator.free(coords);
                    }
                }
                return .{ .type = .Polygon, .coordinates = .{ .Polygon = coords }, .allocator = allocator };
            }

            pub fn initMultiPolygon(allocator: std.mem.Allocator, coordinates: []const []const []const Position) !Geometry {
                const coords = try allocator.alloc([][]Position, coordinates.len);
                errdefer allocator.free(coords);
                for (coordinates, 0..) |polygon, i| {
                    coords[i] = try allocator.alloc([]Position, polygon.len);
                    errdefer {
                        for (coords[0..i]) |c| {
                            allocator.free(c);
                        }
                        allocator.free(coords);
                    }
                    for (polygon, 0..) |line, j| {
                        coords[i][j] = try allocator.dupe(Position, line);
                        errdefer {
                            for (coords[0..i]) |c| {
                                allocator.free(c);
                            }
                            allocator.free(coords);
                        }
                    }
                }
                return .{ .type = .MultiPolygon, .coordinates = .{ .MultiPolygon = coords }, .allocator = allocator };
            }

            pub fn deinit(self: *Geometry) void {
                switch (self.type) {
                    .Point => {},
                    .MultiPoint => self.allocator.free(self.coordinates.MultiPoint),
                    .LineString => self.allocator.free(self.coordinates.LineString),
                    .MultiLineString => {
                        for (self.coordinates.MultiLineString) |line| {
                            self.allocator.free(line);
                        }
                        self.allocator.free(self.coordinates.MultiLineString);
                    },
                    .Polygon => {
                        for (self.coordinates.Polygon) |line| {
                            self.allocator.free(line);
                        }
                        self.allocator.free(self.coordinates.Polygon);
                    },
                    .MultiPolygon => {
                        for (self.coordinates.MultiPolygon) |polygon| {
                            for (polygon) |line| {
                                self.allocator.free(line);
                            }
                            self.allocator.free(polygon);
                        }
                        self.allocator.free(self.coordinates.MultiPolygon);
                    },
                }
            }

            pub fn jsonStringify(self: *const Geometry, ws: anytype) !void {
                try ws.beginObject();

                try ws.objectField("type");
                try ws.write(@tagName(self.type));

                try ws.objectField("coordinates");
                switch (self.coordinates) {
                    .Point => |geometry| try ws.write(geometry),
                    .MultiPoint, .LineString => |geometry| try ws.write(geometry),
                    .MultiLineString, .Polygon => |geometry| try ws.write(geometry),
                    .MultiPolygon => |geometry| try ws.write(geometry),
                }

                try ws.endObject();
            }

            pub fn jsonParse(allocator: std.mem.Allocator, source: anytype, options: std.json.ParseOptions) !@This() {
                var arena = std.heap.ArenaAllocator.init(allocator);
                defer arena.deinit();
                const value = try std.json.parseFromTokenSourceLeaky(std.json.Value, arena.allocator(), source, options);
                return try jsonParseFromValue(allocator, value, options);
            }

            pub fn jsonParseFromValue(allocator: std.mem.Allocator, source: std.json.Value, options: std.json.ParseOptions) !@This() {
                const coords = source.object.get("coordinates") orelse return error.SyntaxError;
                const geometry_type = blk: {
                    const root = source.object;
                    const type_str = root.get("type") orelse return error.SyntaxError;
                    const type_value = type_str.string;
                    const geom_type = std.meta.stringToEnum(GeometryType, type_value) orelse return error.SyntaxError;
                    break :blk geom_type;
                };
                return switch (geometry_type) {
                    .Point => Geometry.initPoint((try std.json.parseFromValue(Position, allocator, coords, options)).value),
                    .MultiPoint => Geometry.initMultiPoint(allocator, (try std.json.parseFromValue([]Position, allocator, coords, options)).value),
                    .LineString => Geometry.initLineString(allocator, (try std.json.parseFromValue([]Position, allocator, coords, options)).value),
                    .MultiLineString => Geometry.initMultiLineString(allocator, (try std.json.parseFromValue([][]Position, allocator, coords, options)).value),
                    .Polygon => Geometry.initPolygon(allocator, (try std.json.parseFromValue([][]Position, allocator, coords, options)).value),
                    .MultiPolygon => Geometry.initMultiPolygon(allocator, (try std.json.parseFromValue([][][]Position, allocator, coords, options)).value),
                };
            }
        };

        pub const Feature = struct {
            type: GeoJsonType = .Feature,
            geometry: Geometry,

            pub fn init(geometry: Geometry) Feature {
                return .{ .geometry = geometry };
            }

            pub fn deinit(self: *Feature) void {
                self.geometry.deinit();
            }

            pub fn jsonStringify(self: *const Feature, ws: anytype) !void {
                try ws.beginObject();

                try ws.objectField("type");
                try ws.write(@tagName(self.type));

                try ws.objectField("geometry");
                try ws.write(self.geometry);

                try ws.endObject();
            }

            pub fn jsonParse(allocator: std.mem.Allocator, source: anytype, options: std.json.ParseOptions) !@This() {
                var arena = std.heap.ArenaAllocator.init(allocator);
                defer arena.deinit();
                const value = try std.json.parseFromTokenSourceLeaky(std.json.Value, arena.allocator(), source, options);
                return try jsonParseFromValue(allocator, value, options);
            }

            pub fn jsonParseFromValue(allocator: std.mem.Allocator, source: std.json.Value, options: std.json.ParseOptions) !@This() {
                const root = source.object;
                const geometry_value = root.get("geometry") orelse return error.SyntaxError;
                const geometry = try Geometry.jsonParseFromValue(allocator, geometry_value, options);
                return Feature.init(geometry);
            }
        };

        pub const FeatureCollection = struct {
            type: GeoJsonType = .FeatureCollection,
            features: std.ArrayList(Feature),

            pub fn init(allocator: std.mem.Allocator) FeatureCollection {
                return .{ .features = std.ArrayList(Feature).init(allocator) };
            }

            pub fn deinit(self: *FeatureCollection) void {
                for (self.features.items) |*feature| {
                    feature.deinit();
                }
                self.features.deinit();
            }

            pub fn jsonStringify(self: *const FeatureCollection, ws: anytype) !void {
                try ws.beginObject();

                try ws.objectField("type");
                try ws.write(@tagName(self.type));

                try ws.objectField("features");
                try ws.beginArray();
                for (self.features.items) |feature| {
                    try ws.write(feature);
                }
                try ws.endArray();

                try ws.endObject();
            }

            pub fn jsonParse(allocator: std.mem.Allocator, source: anytype, options: std.json.ParseOptions) !@This() {
                var arena = std.heap.ArenaAllocator.init(allocator);
                defer arena.deinit();
                const value = try std.json.parseFromTokenSourceLeaky(std.json.Value, arena.allocator(), source, options);
                return try jsonParseFromValue(allocator, value, options);
            }

            pub fn jsonParseFromValue(allocator: std.mem.Allocator, source: std.json.Value, options: std.json.ParseOptions) !@This() {
                const features_value = source.object.get("features") orelse return error.SyntaxError;
                var fc = FeatureCollection.init(allocator);
                for (features_value.array.items) |item| {
                    const feature = try Feature.jsonParseFromValue(allocator, item, options);
                    try fc.features.append(feature);
                }
                return fc;
            }
        };
    };
}

test {
    _ = @import("geojson_test.zig");
}
